@@ -400,6 +400,7 @@ static int nvme_uring_cmd_admin_passthru_async(struct io_uring *ring, struct nvm
 	sqe->fd = args->fd;
 	sqe->opcode = IORING_OP_URING_CMD;
 	sqe->cmd_op = NVME_URING_CMD_ADMIN;
+	sqe->user_data = (__u64)(uintptr_t)args;
 
 	ret = io_uring_submit(ring);
 	if (ret < 0) {
@@ -412,8 +413,9 @@ static int nvme_uring_cmd_admin_passthru_async(struct io_uring *ring, struct nvm
 
 static int nvme_uring_cmd_wait_complete(struct io_uring *ring, int n)
 {
+	struct nvme_get_log_args *args;
 	struct io_uring_cqe *cqe;
-	int i, ret;
+	int i, ret = 0;
 
 	for (i = 0; i < n; i++) {
 		ret = io_uring_wait_cqe(ring, &cqe);
@@ -422,10 +424,18 @@ static int nvme_uring_cmd_wait_complete(struct io_uring *ring, int n)
 			return -1;
 		}
 
+		if (cqe->res) {
+			args = (struct nvme_get_log_args *)cqe->user_data;
+			if (args->result)
+				*args->result = cqe->res;
+			ret = cqe->res;
+			break;
+		}
+
 		io_uring_cqe_seen(ring, cqe);
 	}
 
-	return n;
+	return ret;
 }
 #endif
 
@@ -469,11 +479,14 @@ int nvme_get_log_page(int fd, __u32 xfer_len, struct nvme_get_log_args *args)
 #ifdef CONFIG_LIBURING
 		if (io_uring_kernel_support == IO_URING_AVAILABLE) {
 			if (n >= NVME_URING_ENTRIES) {
-				nvme_uring_cmd_wait_complete(&ring, n);
+				ret = nvme_uring_cmd_wait_complete(&ring, n);
 				n = 0;
 			}
 			n += 1;
 			ret = nvme_uring_cmd_admin_passthru_async(&ring, args);
+
+			if (ret)
+				nvme_uring_cmd_exit(&ring);
 		} else
 #endif
 		ret = nvme_get_log(args);
@@ -488,8 +501,8 @@ int nvme_get_log_page(int fd, __u32 xfer_len, struct nvme_get_log_args *args)
 	if (io_uring_kernel_support == IO_URING_AVAILABLE) {
 		ret = nvme_uring_cmd_wait_complete(&ring, n);
 		nvme_uring_cmd_exit(&ring);
-		if (ret < 0)
-			return -1;
+		if (ret)
+			return ret;
 	}
 #endif
 	return 0;
@@ -709,6 +722,19 @@ int nvme_set_features_temp_thresh(int fd, __u16 tmpth, __u8 tmpsel,
 				   result);
 }
 
+int nvme_set_features_temp_thresh2(int fd, __u16 tmpth, __u8 tmpsel,
+				   enum nvme_feat_tmpthresh_thsel thsel, __u8 tmpthh,
+				   bool save, __u32 *result)
+{
+	__u32 value = NVME_SET(tmpth, FEAT_TT_TMPTH) |
+			NVME_SET(tmpsel, FEAT_TT_TMPSEL) |
+			NVME_SET(thsel, FEAT_TT_THSEL) |
+			NVME_SET(tmpthh, FEAT_TT_TMPTHH);
+
+	return __nvme_set_features(fd, NVME_FEAT_FID_TEMP_THRESH, value, save,
+				   result);
+}
+
 int nvme_set_features_err_recovery(int fd, __u32 nsid, __u16 tler, bool dulbe,
 				   bool save, __u32 *result)
 {
@@ -1065,6 +1091,26 @@ int nvme_get_features_temp_thresh(int fd, enum nvme_get_features_sel sel,
 	return __nvme_get_features(fd, NVME_FEAT_FID_TEMP_THRESH, sel, result);
 }
 
+int nvme_get_features_temp_thresh2(int fd, enum nvme_get_features_sel sel, __u8 tmpsel,
+				   enum nvme_feat_tmpthresh_thsel thsel, __u32 *result)
+{
+	struct nvme_get_features_args args = {
+		.args_size = sizeof(args),
+		.fd = fd,
+		.fid = NVME_FEAT_FID_TEMP_THRESH,
+		.nsid = NVME_NSID_NONE,
+		.sel = sel,
+		.cdw11 = NVME_SET(tmpsel, FEAT_TT_TMPSEL) | NVME_SET(thsel, FEAT_TT_THSEL),
+		.uuidx = NVME_UUID_NONE,
+		.data_len = 0,
+		.data = NULL,
+		.timeout = NVME_DEFAULT_IOCTL_TIMEOUT,
+		.result = result,
+	};
+
+	return nvme_get_features(&args);
+}
+
 int nvme_get_features_err_recovery(int fd, enum nvme_get_features_sel sel,
 				   __u32 *result)
 {
@@ -1920,9 +1966,9 @@ static int nvme_set_var_size_tags(__u32 *cmd_dw2, __u32 *cmd_dw3, __u32 *cmd_dw1
 		return -1;
 	}
 
-	*cmd_dw2 = cpu_to_be32(cdw2);
-	*cmd_dw3 = cpu_to_be32(cdw3);
-	*cmd_dw14 = cpu_to_be32(cdw14);
+	*cmd_dw2 = cdw2;
+	*cmd_dw3 = cdw3;
+	*cmd_dw14 = cdw14;
 	return 0;
 }
 
@@ -2307,15 +2353,27 @@ int nvme_dim_send(struct nvme_dim_args *args)
 
 int nvme_lm_cdq(struct nvme_lm_cdq_args *args)
 {
+	const size_t size_v1 = sizeof_args(struct nvme_lm_cdq_args, sz_u8, __u64);
+	const size_t size_v2 = sizeof_args(struct nvme_lm_cdq_args, sz, __u64);
 	__u32 cdw10 = NVME_SET(args->sel, LM_CDQ_SEL) |
 		      NVME_SET(args->mos, LM_CDQ_MOS);
-	__u32 cdw11 = 0, data_len = 0;
+	__u32 cdw11 = 0, data_len = 0, sz = 0;
 	int err;
 
+	if (args->args_size < size_v1 || args->args_size > size_v2) {
+		errno = EINVAL;
+		return -1;
+	}
+
+	if (args->args_size == size_v1)
+		sz = args->sz_u8;
+	else
+		sz = args->sz;
+
 	if (args->sel == NVME_LM_SEL_CREATE_CDQ) {
 		cdw11 = NVME_SET(args->cntlid, LM_CREATE_CDQ_CNTLID) |
 			NVME_LM_CREATE_CDQ_PC;
-		data_len = args->sz << 2;
+		data_len = sz << 2;
 	} else if (args->sel == NVME_LM_SEL_DELETE_CDQ) {
 		cdw11 = NVME_SET(args->cdqid, LM_DELETE_CDQ_CDQID);
 	}
@@ -2324,7 +2382,7 @@ int nvme_lm_cdq(struct nvme_lm_cdq_args *args)
 		.opcode = nvme_admin_ctrl_data_queue,
 		.cdw10 = cdw10,
 		.cdw11 = cdw11,
-		.cdw12 = args->sz,
+		.cdw12 = sz,
 		.addr = (__u64)(uintptr_t)args->data,
 		.data_len = data_len,
 		.timeout_ms = args->timeout,
